"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.auth = void 0;
const http_errors_1 = __importDefault(require("http-errors"));
const debug_1 = __importDefault(require("debug"));
const api_1 = require("../../lib/api");
const misc_1 = require("./misc");
const SSO_1 = require("./SSO");
const SCHMWithCookies_1 = require("./schm/SCHMWithCookies");
const SCHMWithSync_1 = require("./schm/SCHMWithSync");
const status_1 = require("./status");
const debug = debug_1.default('node-expose-sspi:auth');
/**
 * Tries to get SSO information from browser. If success, the SSO info
 * is stored under req.sso
 *
 * @export
 * @param {AuthOptions} [options={}]
 * @returns {RequestHandler}
 */
function auth(options = {}) {
    const opts = {
        useActiveDirectory: true,
        useGroups: true,
        useOwner: false,
        useCookies: true,
        groupFilterRegex: '.*',
        allowsGuest: false,
        allowsAnonymousLogon: false,
        useSession: false,
    };
    Object.assign(opts, options);
    let { credential, tsExpiry } = api_1.sspi.AcquireCredentialsHandle({
        packageName: 'Negotiate',
    });
    const checkCredentials = () => {
        if (tsExpiry < new Date()) {
            // renew server credentials
            api_1.sspi.FreeCredentialsHandle(credential);
            const renewed = api_1.sspi.AcquireCredentialsHandle({
                packageName: 'Negotiate',
            });
            credential = renewed.credential;
            tsExpiry = renewed.tsExpiry;
        }
    };
    const schManager = opts.useCookies
        ? new SCHMWithCookies_1.SCHMWithCookies()
        : new SCHMWithSync_1.SCHMWithSync(10000);
    let previousServerContextHandle;
    // returns the node middleware.
    return (req, res, next) => {
        if (opts.useSession) {
            const session = req
                .session;
            if (session === null || session === void 0 ? void 0 : session.sso) {
                session.sso.cached = true;
                req.sso = session.sso;
                next();
                return;
            }
        }
        (async () => {
            var _a, _b, _c, _d;
            const cookieToken = schManager.getCookieToken(req, res);
            debug('cookieToken: ', cookieToken);
            let messageType = 'Unknown';
            try {
                const authorization = req.headers.authorization;
                if (!authorization) {
                    debug('no authorization key in header');
                    res.statusCode = 401;
                    res.setHeader('WWW-Authenticate', 'Negotiate');
                    return res.end();
                }
                if (!authorization.startsWith('Negotiate ')) {
                    res.statusCode = 400;
                    return res.end(`Malformed authentication token: ${authorization}`);
                }
                checkCredentials();
                const token = authorization.substring('Negotiate '.length);
                messageType = misc_1.getMessageType(token);
                debug('messageType: ', messageType);
                const buffer = misc_1.decode(token);
                debug(misc_1.hexDump(buffer));
                // test if first token
                if (messageType === 'NTLM_NEGOTIATE_01' ||
                    messageType === 'Kerberos_1') {
                    await schManager.waitForReleased(cookieToken);
                    debug('schManager waitForReleased finished.');
                    const ssoMethod = messageType.startsWith('NTLM')
                        ? 'NTLM'
                        : 'Kerberos';
                    schManager.setMethod(ssoMethod, cookieToken);
                }
                const input = {
                    credential,
                    SecBufferDesc: {
                        ulVersion: 0,
                        buffers: [buffer],
                    },
                };
                const serverContextHandle = schManager.getHandle(cookieToken);
                if (serverContextHandle) {
                    debug('adding to input a serverContextHandle (not first exchange)');
                    input.contextHandle = serverContextHandle;
                }
                if (!serverContextHandle && messageType !== 'NTLM_NEGOTIATE_01') {
                    debug('set cookie bug management');
                    input.contextHandle = previousServerContextHandle;
                }
                debug('input just before calling AcceptSecurityContext', input);
                const serverSecurityContext = api_1.sspi.AcceptSecurityContext(input);
                debug('serverSecurityContext just after AcceptSecurityContext', serverSecurityContext);
                if (!['SEC_E_OK', 'SEC_I_CONTINUE_NEEDED'].includes(serverSecurityContext.SECURITY_STATUS)) {
                    // 'SEC_I_COMPLETE_AND_CONTINUE', 'SEC_I_COMPLETE_NEEDED' are considered as errors because it is used
                    // only by 'Digest' SSP. (not by Negotiate, Kerberos or NTLM)
                    if (serverSecurityContext.SECURITY_STATUS === 'SEC_E_LOGON_DENIED') {
                        schManager.release(cookieToken);
                        next(http_errors_1.default(401, `SEC_E_LOGON_DENIED. (incorrect login/password, or account disabled, or locked, etc.). Protocol Message = ${messageType}.`));
                        return;
                    }
                    throw new Error('AcceptSecurityContext error: ' +
                        serverSecurityContext.SECURITY_STATUS);
                }
                schManager.setHandle(serverSecurityContext.contextHandle, cookieToken);
                previousServerContextHandle = serverSecurityContext.contextHandle;
                debug('AcceptSecurityContext output buffer');
                debug(misc_1.hexDump(serverSecurityContext.SecBufferDesc.buffers[0]));
                if (serverSecurityContext.SECURITY_STATUS === 'SEC_I_CONTINUE_NEEDED') {
                    res.statusCode = 401;
                    res.setHeader('WWW-Authenticate', 'Negotiate ' +
                        misc_1.encode(serverSecurityContext.SecBufferDesc.buffers[0]));
                    return res.end();
                }
                const lastServerContextHandle = schManager.getHandle(cookieToken);
                if (!lastServerContextHandle) {
                    throw new Error('cannot get the server context handle');
                }
                const method = schManager.getMethod(cookieToken);
                const sso = new SSO_1.SSO(lastServerContextHandle, method);
                sso.setOptions(opts);
                await sso.load();
                req.sso = sso.getJSON();
                if (opts.useSession) {
                    const session = req
                        .session;
                    if (session) {
                        req.sso.cached = false;
                        session.sso = req.sso;
                    }
                }
                api_1.sspi.DeleteSecurityContext(lastServerContextHandle);
                schManager.release(cookieToken);
                // check if user is allowed.
                if (!opts.allowsAnonymousLogon &&
                    ((_b = (_a = req.sso) === null || _a === void 0 ? void 0 : _a.user) === null || _b === void 0 ? void 0 : _b.name) === 'ANONYMOUS LOGON') {
                    res.statusCode = 401;
                    return res.end('Anonymous login not authorized.');
                }
                if (!opts.allowsGuest && ((_d = (_c = req.sso) === null || _c === void 0 ? void 0 : _c.user) === null || _d === void 0 ? void 0 : _d.name) === 'Guest') {
                    res.statusCode = 401;
                    return res.end('Guest not authorized.');
                }
                // user authenticated and allowed.
                res.setHeader('WWW-Authenticate', 'Negotiate ' + misc_1.encode(serverSecurityContext.SecBufferDesc.buffers[0]));
                return next();
            }
            catch (e) {
                schManager.release(cookieToken);
                console.error(e);
                console.error('statusInfo: ', status_1.getStatusInfo());
                console.error('messageType: ', messageType);
                next(http_errors_1.default(401, `Error while doing SSO: ${e.message}`));
            }
        })();
    };
}
exports.auth = auth;
//# sourceMappingURL=auth.js.map